
mat3 rotx(float a){float s = sin(a);float c = cos(a);return mat3(vec3(1.0, 0.0, 0.0), vec3(0.0, c, s), vec3(0.0, -s, c));  }
mat3 roty(float a){float s = sin(a);float c = cos(a);return mat3(vec3(c, 0.0, s), vec3(0.0, 1.0, 0.0), vec3(-s, 0.0, c));}
mat3 rotz(float a){float s = sin(a);float c = cos(a);return mat3(vec3(c, s, 0.0), vec3(-s, c, 0.0), vec3(0.0, 0.0, 1.0 ));}


// --------------

// return rd in Equirectangular projection
// uv is screen space
void panorama_screen_uv_to_rd(vec2 uv, out vec3 rd){
    float M_PI = 3.1415926535;
    float ymul = 2.0; float ydiff = -1.0;
    uv.x = 2.0 * uv.x - 1.0;
    uv.y = ymul * uv.y + ydiff;
    rd=vec3(0.0, 0.0, 1.0)*rotx(-uv.y*M_PI/2.0)*roty(-uv.x*M_PI);
    if(length(rd)<0.0001)rd+=0.0001;
    rd = normalize(rd);
}

// return uv for panorama texture
// rd is cubemap rayDir
void panorama_rd_to_pan_uv(vec3 rd, out vec2 uv){
    if(length(rd)<0.0001)rd+=0.0001;
    rd=normalize(rd);
    float M_PI = 3.1415926535;
    float theta = atan(rd.z,rd.x);
    float r = max(length(rd),0.0001);
    float phi =  asin(rd.y/r);
    uv = 0.5 + vec2(theta/(M_PI*2.), -phi/M_PI);
}

// --------------


// --------------
// https://en.wikipedia.org/wiki/Cube_mapping

void convert_xyz_to_cube_uv(in vec3 rd, out int index, out vec2 uv)
{
    int Front_side = 0;
    int Back_side = 1;
    int Up_side = 2;
    int Down_side = 3;
    int Left_side = 4;
    int Right_side = 5;
    
    uv = vec2(0.);
    index = -1;
    if(length(rd)<0.0001)return;
    
    vec3 a = abs(rd);
    bvec3 ip = greaterThan(rd,vec3(0.));
    vec2 uvc = vec2(0.);
    if (ip.x && a.x >= a.y && a.x >= a.z) {uvc.x = -rd.z;uvc.y = rd.y;
    uv = vec2(0.5 * (uvc / a.x + 1.)); index = Front_side;return;
    }else
    if (!ip.x && a.x >= a.y && a.x >= a.z) {uvc.x = rd.z;uvc.y = rd.y;
    uv = vec2(0.5 * (uvc / a.x + 1.)); index = Back_side;return;
    }else
    if (ip.y && a.y >= a.x && a.y >= a.z) {uvc.x = rd.x;uvc.y = -rd.z;
    uv = vec2(0.5 * (uvc / a.y + 1.)); index = Up_side;return;
    }else
    if (!ip.y && a.y >= a.x && a.y >= a.z) {uvc.x = rd.x;uvc.y = rd.z;
    uv = vec2(0.5 * (uvc / a.y + 1.)); index = Down_side;return;
    }else
    if (ip.z && a.z >= a.x && a.z >= a.y) {uvc.x = rd.x;uvc.y = rd.y;
    uv = vec2(0.5 * (uvc / a.z + 1.)); index = Left_side;return;
    }else
    if (!ip.z && a.z >= a.x && a.z >= a.y) {uvc.x = -rd.x;uvc.y = rd.y;
    uv = vec2(0.5 * (uvc / a.z + 1.)); index = Right_side;return;
    }
}

void convert_cube_uv_to_xyz(in int index, in vec2 uv, out vec3 rd)
{
  float uc = 2.0 * uv.x - 1.0;
  float vc = 2.0 * uv.y - 1.0;
  rd=vec3(0.0001,0.9999,0.0001);
  if(index==0){rd=normalize(vec3(1.0, vc, -uc));return;}
  if(index==1){rd=normalize(vec3(-1.0, vc, uc));return;}
  if(index==2){rd=normalize(vec3(uc, 1.0, -vc));return;}
  if(index==3){rd=normalize(vec3(uc,-1.0, vc));return;}
  if(index==4){rd=normalize(vec3(uc, vc, 1.0));return;}
  if(index==5){rd=normalize(vec3(-uc, vc, -1.0));return;}
}

// --------------


// --------------
//cubemap emulation in buffer

// fragCoord and fbo_resolution is screen space buffer coord and iResolution
// return rd, fc_side-fragCoord for cube side, index-side index (in Wikipedia cubemap order)
// side_size-size of cube side in channel
void buffer_cubemap_rd(in vec2 fragCoord, out vec3 rd, out vec2 fc_side, out int index,
                        out int side_size, vec2 fbo_resolution){
    fc_side = vec2(0.5);
    rd = vec3(0.0001, 0.9999, 0.0001);
    index = -1;
    
    ivec2 ipx = ivec2(fragCoord.xy);
    side_size = min(int(fbo_resolution.x)/3,int(fbo_resolution.y)/2);
    bvec2 box = lessThan(ipx/side_size,ivec2(3,2));
    
    if(!(box.x&&box.y))return;
    
    fc_side = vec2(ipx%side_size)+0.5001;
    index = ipx.x/side_size + (ipx.y/side_size)*3;
    
    convert_cube_uv_to_xyz(index, fc_side/vec2(side_size), rd);
}

// return color of buffered cubemap
// Fixed interpolation issue https://www.shadertoy.com/view/sdBfzw
// if you need exact fragCoord on cube side look code after this function
vec4 texture_buffer_cubemap(sampler2D c, vec3 rd){
    int index; vec2 side_uv;
    convert_xyz_to_cube_uv(rd, index, side_uv);
    
    ivec2 shift=ivec2(0);
    if(index==0)shift=ivec2(0, 0); else
    if(index==1)shift=ivec2(1, 0); else
    if(index==2)shift=ivec2(2, 0); else
    if(index==3)shift=ivec2(0, 1); else
    if(index==4)shift=ivec2(1, 1); else
    if(index==5)shift=ivec2(2, 1);
    
    
    vec2 s = vec2(1./3.,1./2.);
    
    // fixed interpolation
    
    vec2 fbo_resolution = vec2(textureSize(c, 0).xy);
    int side_size = min(int(fbo_resolution.x)/3,int(fbo_resolution.y)/2);
    vec2 tres = float(side_size)*(1./s);
    vec2 prop = (tres/fbo_resolution)*s;
    
    vec2 uv = side_uv*prop + vec2(shift)*prop;
    
    return textureLod(c, uv, 0.);
}

/*
// Warning - texture interpolation may/will be broken if you use this texture_buffer_cubemap
// use this when you need exact fragCoord of cubemap
vec4 texture_buffer_cubemap(sampler2D c, vec3 rd){
    int index; vec2 side_uv;
    convert_xyz_to_cube_uv(rd, index, side_uv);
    vec2 fbo_resolution = vec2(textureSize(c, 0).xy);
    int side_size = min(int(fbo_resolution.x)/3,int(fbo_resolution.y)/2);
    
    ivec2 shift=ivec2(0);
    if(index==0)shift=ivec2(0, 0); else
    if(index==1)shift=ivec2(1, 0); else
    if(index==2)shift=ivec2(2, 0); else
    if(index==3)shift=ivec2(0, 1); else
    if(index==4)shift=ivec2(1, 1); else
    if(index==5)shift=ivec2(2, 1);
    
    
    vec2 s = float(side_size)/fbo_resolution;
    
    // Warning - texture interpolation may be broken if you use next line
    vec2 fc = floor(side_uv*float(side_size))+0.5; //BUG may ignore texture interpolaton BUT correct fragCoord fc value
    //vec2 fc = (side_uv*float(side_size))+0.; // BAD fix
    
    vec2 uv = (fc/float(side_size))*s + vec2(shift*side_size)/fbo_resolution;
    
    return textureLod(c, uv, 0.);
}
*/
// --------------


// --------------

//https://www.shadertoy.com/view/tdjXDt

// return rd to use as cubemap coordinate
// inUV is screen uv
vec3 getCubeMap(vec2 inUV)
{
  vec3 samplePos = vec3(0.0001,0.9999,0.0001);
  
  // Crude statement to visualize different cube map faces based on UV coordinates
  int x = int(floor(inUV.x / 0.25f));
  int y = int(floor(inUV.y / (1.0 / 3.0))); 
  if (y == 1) {
    vec2 uv = vec2(inUV.x * 4.0f, (inUV.y - 1.0/3.0) * 3.0);
    uv = 2.0 * vec2(uv.x - float(x) * 1.0, uv.y) - 1.0;
    switch (x) {
      case 0: // NEGATIVE_X
        samplePos = vec3(-1.0f, uv.y, uv.x);
        break;
      case 1: // POSITIVE_Z       
        samplePos = vec3(uv.x, uv.y, 1.0f);
        break;
      case 2: // POSITIVE_X
        samplePos = vec3(1.0, uv.y, -uv.x);
        break;        
      case 3: // NEGATIVE_Z
        samplePos = vec3(-uv.x, uv.y, -1.0f);
        break;
    }
  } else {
    if (x == 1) { 
      vec2 uv = vec2((inUV.x - 0.25) * 4.0, (inUV.y - float(y) / 3.0) * 3.0);
      uv = 2.0 * uv - 1.0;
      switch (y) {
        case 0: // NEGATIVE_Y
          samplePos = vec3(uv.x, -1.0f, uv.y);
          break;
        case 2: // POSITIVE_Y
          samplePos = vec3(uv.x, 1.0f, -uv.y);
          break;
      }
    }
  }

  return normalize(samplePos);  
}

// --------------


// --------------
// lines in rd, p is rd

float dist_vertical_line( vec3 p)
{
  return length(vec2(length(p.xz)-1.0,p.y));
}

float dist_horizontal_line( vec3 p)
{
  return length(vec2(length(p.xy)-1.0,p.z));
}

// --------------




















// --------------
// debug drawing 2D, not related to Cubemap logic


//https://www.shadertoy.com/view/ldsyz4
// The MIT License
// Copyright © 2017 Inigo Quilez
// Digit data by P_Malin (https://www.shadertoy.com/view/4sf3RN)
const int[] font = int[](0x75557, 0x22222, 0x74717, 0x74747, 0x11574, 0x71747, 0x71757, 0x74444, 0x75757, 0x75747);
const int[] powers = int[](1, 10, 100, 1000, 10000, 100000, 1000000);

int PrintInt( in vec2 uv, in int value, const int maxDigits )
{
    if( abs(uv.y-0.5)<0.5 )
    {
        int iu = int(floor(uv.x));
        if( iu>=0 && iu<maxDigits )
        {
            int n = (value/powers[maxDigits-iu-1]) % 10;
            uv.x = fract(uv.x);//(uv.x-float(iu)); 
            ivec2 p = ivec2(floor(uv*vec2(4.0,5.0)));
            return (font[n] >> (p.x+p.y*4)) & 1;
        }
    }
    return 0;
}


// https://www.shadertoy.com/view/XsBBRd

// ------------------------------------------------------------------------------------------------
float m_stretch(float point, float stretch){
    return .5 * (sign(point) * stretch - point) * (sign(abs(point) - stretch) + 1.);
}

float ollj_rotate(vec2 uv){
    const float ROTATE_PARAM0 = sqrt(1.);
    const float ROTATE_PARAM1 = sqrt(.0);
    return dot(uv, vec2(ROTATE_PARAM0 + ROTATE_PARAM1, ROTATE_PARAM0 - ROTATE_PARAM1));
}


float sdf_arrow(vec2 uv, float norm, float angle, float head_height, float stem_width){
    uv = vec2(cos(angle) * uv.x + sin(angle) * uv.y, -sin(angle) * uv.x + cos(angle) * uv.y);

    norm -= head_height;
    uv.x -= norm;

    uv.y = abs(uv.y);   
    float head = max(ollj_rotate(uv) - head_height, -uv.x);

    uv.x = (.5 * m_stretch(2. * uv.x + norm, norm));
    uv.y = (.5 * m_stretch(2. * uv.y, stem_width));
    float stem = length(uv);

    return min(head, stem);
}

float get_arrow(vec2 arrow_uv, float arrow_aa, float arrow_angle){
    float arrow_norm = 1.15; 
    float arrow_head_height = .1;  
    float arrow_stem_width = .04; 
    float arrow;
    arrow = sdf_arrow(arrow_uv, arrow_norm, arrow_angle, arrow_head_height, arrow_stem_width);
    arrow = 1.-smoothstep(0., arrow_aa, arrow);

    return arrow;
}

// ------------------------------------------------------------------------------------------------


vec4 grid(vec2 p, float grid_line, int id, vec3 idx_color){
    vec3 colr_r=vec3(0.8,0.,0.);
    vec3 colr_g=vec3(0.0,0.8,0.);
    vec3 colr_r2=vec3(0.35,0.,0.);
    vec3 colr_gr=vec3(0.16);
    vec3 colr_w=vec3(0.98);
    const float grid_size=2.;
    
    vec2 op=p;
    
    vec3 col=vec3(0.);
    float al=0.;
    vec2 dv=1.-step(0.5,abs(p));
    if(dv.x*dv.y==0.)return vec4(col,al);
    p=fract(p*grid_size)-0.5;
    dv=step(0.35,abs(p));
    if(dv.x*dv.y==0.){
        dv=step(0.5-grid_line*grid_size,abs(p));
        float r1=max(dv.x,dv.y);
        p=fract(p*2.)-0.5;
        dv=step(0.5-grid_line*grid_size*2.,abs(p));
        float r2=max(dv.x,dv.y);
        p=fract((p-0.5)*5.)-0.5;
        dv=step(0.5-grid_line*grid_size*2.*5.,abs(p));
        float gr=max(dv.x,dv.y);
        col=colr_r*r1;
        col=max(col,colr_r2*r2);
        col=max(col,colr_gr*gr*(1.-r1));
        al=max(max(r1,r2),gr);
    }
    else{
        dv=step(0.4,abs(p));
        if(dv.x*dv.y>0.){
            dv=step(0.5-grid_line*grid_size*2.,abs(p));
            al=max(dv.x,dv.y);
            col=colr_w*al;
        }
    }
    float arow = get_arrow(op*2.+vec2(0.25,0.45), grid_line*4., 0.);
    al=max(al, arow);
    col=mix(col, colr_g*0.7, arow);
    
    arow = get_arrow(op*2.+vec2(0.45,0.25), grid_line*4., 3.1415926/2.);
    al=max(al, arow);
    col=mix(col, colr_r*0.7, arow);
    
    float idx=float(PrintInt((op+vec2(-0.25))*6.,id,1));
    al=max(al, idx);
    col=mix(col, idx_color, idx);
    
    return vec4(col,al);
}

// --------------




