#version 450

#extension GL_ARB_separate_shader_objects: enable
#extension GL_ARB_shading_language_420pack: enable
#extension GL_GOOGLE_include_directive : enable

layout (location = 0) in vec2 frag_pos;

#define OFFSCREEN_BUFFERS 4
#define CUBEMAP_BUFFER 1
#define IMAGE_TEXTURES 4

layout (set = 0, binding = 0) uniform sampler2D iTextures[IMAGE_TEXTURES]; //textures
layout (set = 0, binding = IMAGE_TEXTURES) uniform sampler2D iChannels[(OFFSCREEN_BUFFERS + CUBEMAP_BUFFER)]; //buffers
layout (set = 0, binding = IMAGE_TEXTURES+(OFFSCREEN_BUFFERS + CUBEMAP_BUFFER)) uniform sampler2D iKeyboard; //keyboard

//define in any order
#define iChannel0 iChannels[0]
#define iChannel1 iChannels[1]
#define iChannel2 iChannels[2]
#define iChannel3 iChannels[3]
#define iCubemap iChannels[(OFFSCREEN_BUFFERS + CUBEMAP_BUFFER)-1]

layout (push_constant) uniform push_constants
{
    vec4 u_Mouse;
    vec4 u_Date;
    bvec2 u_Mouse_lr;
    vec2 u_Resolution;
    bool u_debugdraw;
    int pCustom;
    float u_Time;
    float u_TimeDelta;
    int u_Frame;
} constants;

vec3 iResolution=vec3(vec2(1024.),1.);
float iTime=constants.u_Time;
float iTimeDelta=constants.u_TimeDelta;
int iFrame=constants.u_Frame;
vec4 iMouse=constants.u_Mouse;
vec4 iDate=constants.u_Date;
bool is_debugdraw=constants.u_debugdraw;
bool is_pause=bool(constants.pCustom-(constants.pCustom/10)*10);

layout (location = 0) out vec4 out_color;

#include "../shadertoy/common.glsl"
#include "../shadertoy/cube_buf.glsl"

void buffer_cubemap_rd_local(in vec2 fragCoord, out vec3 rd, out vec2 fc_side, out int index,
                        out int side_size, vec2 fbo_resolution);
void main()
{
    vec4 uFragColor=vec4(0.);
    out_color=uFragColor;
    vec2 fragCoord=gl_FragCoord.xy;
    
    vec3 rd;vec2 fc_cube;int index;int side_size;
    buffer_cubemap_rd_local(fragCoord, rd, fc_cube, index, side_size, vec2(3.,2.)*1024.);
    if(index<0)return;
    fc_cube.y = float(side_size)-fc_cube.y-0.0001;
    
    mainCubemap(uFragColor, fc_cube, vec3(0.0), rd);
    out_color=uFragColor;
}

void convert_cube_uv_to_xyz_local(in int index, in vec2 uv, out vec3 rd)
{
  float uc = 2.0 * uv.x - 1.0;
  float vc = 2.0 * uv.y - 1.0;
  rd=vec3(0.0001,0.9999,0.0001);
  if(index==0){rd=normalize(vec3(1.0, vc, -uc));return;}
  if(index==1){rd=normalize(vec3(-1.0, vc, uc));return;}
  if(index==2){rd=normalize(vec3(uc, 1.0, -vc));return;}
  if(index==3){rd=normalize(vec3(uc,-1.0, vc));return;}
  if(index==4){rd=normalize(vec3(uc, vc, 1.0));return;}
  if(index==5){rd=normalize(vec3(-uc, vc, -1.0));return;}
}


void buffer_cubemap_rd_local(in vec2 fragCoord, out vec3 rd, out vec2 fc_side, out int index,
                        out int side_size, vec2 fbo_resolution){
    fc_side = vec2(0.5);
    rd = vec3(0.0001, 0.9999, 0.0001);
    index = -1;
    
    ivec2 ipx = ivec2(fragCoord.xy);
    side_size = min(int(fbo_resolution.x)/3,int(fbo_resolution.y)/2);
    bvec2 box = lessThan(ipx/side_size,ivec2(3,2));
    
    if(!(box.x&&box.y))return;
    
    fc_side = vec2(ipx%side_size)+0.5001;
    index = ipx.x/side_size + (ipx.y/side_size)*3;
    
    convert_cube_uv_to_xyz_local(index, fc_side/vec2(side_size), rd);
}

